"""
_GetCompletedBlocks_

MySQL implementation of RucioInjector.GetClompletedBlocks

Retrieve a list of blocks that are complete,
including their location and the sites they
are subscribed too

"""


from WMCore.Database.DBFormatter import DBFormatter


class GetCompletedBlocks(DBFormatter):
    """
    Retrieves a list of blocks that are closed but NOT sure yet if they are deleteable:
      - The workflows for all files in the block need to be completed (This relates only to the
        workflows directly producing the files and does not track child workflows completion)
      - The subscription made for dataset is copy+delete
      - A subscription has been made at the Data Management system
      - The blocks hasn't been deleted yet
    """
    sql = """SELECT dbsbuffer_block.blockname,
                    dbsbuffer_location.pnn,
                    dbsbuffer_dataset.path,
                    dbsbuffer_dataset_subscription.site,
                    dbsbuffer_workflow.name,
                    dbsbuffer_block.create_time
             FROM dbsbuffer_dataset_subscription
             INNER JOIN dbsbuffer_dataset ON
               dbsbuffer_dataset.id = dbsbuffer_dataset_subscription.dataset_id
             INNER JOIN dbsbuffer_block ON
               dbsbuffer_block.dataset_id = dbsbuffer_dataset_subscription.dataset_id
             INNER JOIN dbsbuffer_file ON
               dbsbuffer_file.block_id = dbsbuffer_block.id
             INNER JOIN dbsbuffer_workflow ON
               dbsbuffer_workflow.id = dbsbuffer_file.workflow
             INNER JOIN dbsbuffer_location ON
               dbsbuffer_location.id = dbsbuffer_block.location
             WHERE dbsbuffer_dataset_subscription.delete_blocks = 1
               AND dbsbuffer_dataset_subscription.subscribed = 1
               AND dbsbuffer_block.status = 'Closed'
               AND dbsbuffer_block.deleted = 0
             GROUP BY dbsbuffer_block.blockname,
                      dbsbuffer_location.pnn,
                      dbsbuffer_dataset.path,
                      dbsbuffer_dataset_subscription.site,
                      dbsbuffer_workflow.name,
                      dbsbuffer_block.create_time
             HAVING COUNT(*) = SUM(dbsbuffer_workflow.completed)
             """

    def format(self, result):
        """
        _format_

        Format the query results into the proper dictionary expected at the upper layer Python code.
        The input should be a list of database objects each representing a line returned from the database
        with key names matching the column names from the sql query
        The result should be a list of dictionaries one record per line returned from the database
        with key names mapped to the python code expected structures.

        e.g.
        [{'blockName': '/Cosmics/Tier0_REPLAY_2022-v425/RAW#aa3dbb67-dab0-4671-a550-711bdf9f4b08',
          'workflowName': 'Repack_Run351572_StreamPhysics_Tier0_REPLAY_2022_ID220513161632_v425_220514_2038',
          'location': 'T0_CH_CERN_Disk',
          'site': {'T0_CH_CERN_Disk'},
          'dataset': '/Cosmics/Tier0_REPLAY_2022-v425/RAW'},
         {'blockName': '/NoBPTX/Tier0_REPLAY_2022-v425/RAW#6b8e95c3-656b-4302-8265-275df1195f4c',
          'workflowName': 'Repack_Run351572_StreamPhysics_Tier0_REPLAY_2022_ID220513161632_v425_220514_2038',
          'location': 'T0_CH_CERN_Disk',
          'site': {'T0_CH_CERN_Disk'},
          'dataset': '/NoBPTX/Tier0_REPLAY_2022-v425/RAW'}]


        NOTE:
         * location: Means where the output block has been created
         * site(s):  Means where the dataset gets a container-level rule

        :param result: The result as returned by the mysql query execution.
        :return:       List of dictionaries
        """

        # NOTE: We need to rename all the keys to follow the cammelCase standard. And also to comply
        #       with the key names as expected from the rest of the already existing python code
        keyMap = {'blockname': 'blockName',
                  'name': 'workflowName',
                  'pnn': 'location',
                  'site': 'sites',
                  'path': 'dataset',
                  'create_time': 'blockCreateTime'}

        dictResults = DBFormatter.formatDict(self, result)
        for record in dictResults:
            for dbKey, pyKey in keyMap.items():
                if dbKey == 'site':
                    sites = record.pop(dbKey)
                    record[pyKey] = set()
                    record[pyKey].add(sites)
                else:
                    record[pyKey] = record.pop(dbKey)

        return dictResults

    def execute(self, conn=None, transaction=False, returnCursor=False):
        """
        Executing the current sql query.
        :param conn:        A current database connection to be used if existing
        :param transaction: A current database transaction to be used if existing
        :return:            A list of dictionaries one record for each database line returned
        """
        results = self.dbi.processData(self.sql, conn=conn,
                                       transaction=transaction)

        return self.format(results)



"""
_SubscriptionList_

Module with data structures to handle PhEDEx subscriptions
in bulk.
"""
import logging

from WMCore.WMException import WMException

PhEDEx_VALID_SUBSCRIPTION_PRIORITIES = ['low', 'normal', 'high', 'reserved']


class PhEDExSubscriptionException(WMException):
    """
    _PhEDExSubscriptionException_

    Exception class for the phedex subscription
    """
    pass

class PhEDExSubscription(object):
    """
    _PhEDExSubscription_

    Data structure which contains PHEDEx fields for
    PhEDEx subscription data service
    """
    def __init__(self, datasetPathList, nodeList, group, level = 'dataset',
                 priority = 'normal', move = 'n', static = 'n', custodial = 'n',
                 request_only = 'y', blocks = None, subscriptionId = -1, comments=""):
        """
        Initialize PhEDEx subscription with default value
        """
        if isinstance(datasetPathList, basestring):
            datasetPathList = [datasetPathList]
        if isinstance(nodeList, basestring):
            nodeList = [nodeList]

        self.datasetPaths = set(datasetPathList)
        self.nodes = set(nodeList)

        self.level = level.lower()
        self.priority = priority.lower()
        self.move = move.lower()
        self.static = static.lower()
        self.group = group
        self.custodial = custodial.lower()
        self.request_only = request_only.lower()
        self.requesterID = None
        self.status = "New"
        self.comments = comments

        # Subscription id for internal accounting
        self.subscriptionIds = set([subscriptionId])

        # Optional blocks for non-dataset subscriptions
        self.blocks = blocks

        try:
            # Validation checks on the subscription
            for option in (self.static, self.custodial, self.request_only, self.move):
                assert option in ('y', 'n')
            assert self.priority in PhEDEx_VALID_SUBSCRIPTION_PRIORITIES
            assert self.level in ('dataset', 'block')
            if self.level == 'block':
                assert self.blocks is not None
        except AssertionError:
            msg = "The subscription is not a valid PhEDEx subscription.\n"
            msg += "Check the options for this subscription: \n"
            msg += "level: %s\n" % self.level
            msg += "priority: %s\n" % self.priority
            msg += "static: %s\n" % self.static
            msg += "move: %s\n" % self.move
            msg += "custodial: %s\n" % self.custodial
            msg += "blocks: %s\n" % str(self.blocks)
            raise PhEDExSubscriptionException(msg)

    def __str__(self):
        """
        Write out useful information for this object
        :return:
        """
        res = {'datasetPaths': self.datasetPaths, 'nodes': self.nodes,
               'priority': self.priority, 'move': self.move,
               'group': self.group, 'custodial': self.custodial,
               'request_only': self.request_only, 'blocks': self.blocks}
        return str(res)

    def isEqualOptions(self, subscription):
        return (self.level == subscription.level
                and self.priority == subscription.priority
                and self.request_only == subscription.request_only
                and self.custodial == subscription.custodial
                and self.group == subscription.group
                and self.move == subscription.move
                and self.static == subscription.static)

    def isEqualDatasetPaths(self, subscription):
        return (self.datasetPaths == subscription.datasetPaths
                and self.isEqualOptions(subscription))

    def isEqualNode(self, subscription):
        return (self.nodes == subscription.nodes
                and self.isEqualOptions(subscription))

    def addDatasetPaths(self, subscription):
        if self.requesterID != None:
            msg = """ PhEDEx subscription is already made with id: %s\n
                      Create a new subscription
                  """ % (self.requesterID)
            raise Exception(msg)

        self.datasetPaths = self.datasetPaths.union(subscription.datasetPaths)
        self.subscriptionIds = self.subscriptionIds.union(subscription.subscriptionIds)

    def addNodes(self, subscription):
        if self.requesterID != None:
            msg = """ PhEDEx subscription is already made with id: %s\n
                      Create a new subscription
                  """ % (self.requesterID)
            raise Exception(msg)

        self.nodes = self.nodes.union(subscription.nodes)
        self.subscriptionIds = self.subscriptionIds.union(subscription.subscriptionIds)

    def getDatasetPaths(self):
        return list(self.datasetPaths)

    def getSubscriptionIds(self):
        return list(self.subscriptionIds)

    def getDatasetsAndBlocks(self):
        """
        _getDatasetsAndBlocks_

        Get the block structure
        with datasets and blocks
        """
        return self.blocks

    def getBlocks(self):
        """
        _getBlocks_

        Return only the list of blocks
        """
        blocks = []
        for dset, listBlocks in self.blocks.items():
            blocks.extend(listBlocks)
        return blocks

    def getNodes(self):
        return list(self.nodes)

    def getRequesterID(self):
        return self.requesterID

    def setRequesterID(self, requesterId):

        if self.requesterID == None:
            self.requesterID = requesterId
        else:
            msg = """ PhEDEx subscription is already made with id: %s\n
                      Create a new subscription
                  """ % (self.requesterID)
            raise Exception(msg)

    def matchesExistingTransferRequest(self, phedexDataSvc):
        """
        _matchesExistingTransferRequest_

        Check the given phedex data service to verify if an unapproved
        transfer request equal to this subscription is already in the system.
        """
        if len(self.datasetPaths) != 1 or len(self.nodes) != 1:
            msg = "matchesExistingTransferRequest can only run in single node/dataset subscriptions"
            raise PhEDExSubscriptionException(msg)
        if self.level != 'dataset':
            msg = "matchesExistingTransferRequest is only supported by dataset subscriptions"
            raise PhEDExSubscriptionException(msg)

        node = next(iter(self.nodes))
        dataset = next(iter(self.datasetPaths))
        # Get the unapproved requests involving the node and dataset in this subscription
        existingRequests = phedexDataSvc.getRequestList(dataset = dataset,
                                                        node = node,
                                                        decision = 'pending')['phedex']['request']
        for request in existingRequests:
            # Get the detailed information in the request
            requestId = request['id']
            requestInfo = phedexDataSvc.getTransferRequests(request = requestId)['phedex']['request']
            if not requestInfo:
                logging.error("Transfer request %s doesn't exist in PhEDEx", requestId)
                continue # Strange, but let it go.
            requestInfo = requestInfo[0] # It's a singleton
            # Make sure that the node is in the destinations
            destinations = requestInfo['destinations']['node']
            for nodeInfo in destinations:
                if nodeInfo['name'] == node:
                    break
            else:
                continue
            # Create a subscription with this info
            phedexRequest = PhEDExSubscription(self.datasetPaths, self.nodes,
                                               self.group, self.level, requestInfo['priority'],
                                               requestInfo['move'], requestInfo['static'],
                                               requestInfo['custodial'], self.request_only)
            if self.isEqualOptions(phedexRequest):
                return True

        return False

    def matchesExistingSubscription(self, phedexDataSvc):
        """
        _matchesExistingSubscription_

        Check the given phedex data service to verify if a PhEDEx subscription
        equal to this subscription is already in the system.
        """
        if len(self.datasetPaths) != 1 or len(self.nodes) != 1:
            msg = "matchesExistingSubscription can only run in single node/dataset subscriptions"
            raise PhEDExSubscriptionException(msg)
        if self.level != 'dataset':
            msg = "matchesExistingSubscription is only supported by dataset subscriptions"
            raise PhEDExSubscriptionException(msg)

        node = next(iter(self.nodes))
        dataset = next(iter(self.datasetPaths))
        # Check if the dataset has a subscription the given node
        existingSubscription = phedexDataSvc.subscriptions(dataset = dataset,
                                                           node = node)['phedex']['dataset']
        if len(existingSubscription) < 1:
            # No subscriptions
            return False
        datasetInfo = existingSubscription[0]
        for subscriptionInfo in datasetInfo['subscription']:
            # Check that the node in the subscription matches the current node
            if node != subscriptionInfo['node']:
                continue
            # Create a subscription with the info
            phedexSub = PhEDExSubscription(self.datasetPaths, self.nodes,
                                           self.group, subscriptionInfo['level'],
                                           subscriptionInfo['priority'], subscriptionInfo['move'],
                                           self.static, subscriptionInfo['custodial'],
                                           self.request_only)
            if self.isEqualOptions(phedexSub):
                return True

        return False

class SubscriptionList(object):
    """
    _SubscriptionList_

    Class represents collection of subscription.
    This organizes the subscriptions in a way to minimize their number.
    """
    def __init__(self):
        self._subList = []

    def addSubscription(self, subObj):
        """
        _addSubscription_
        Add a new subscription to the subscription policy.
        If the same subscription key exist just add the node list
        """

        for subscription in self._subList:
            if subscription.isEqualOptions(subObj):
                if subscription.isEqualNode(subObj):
                    subscription.addDatasetPaths(subObj)
                    return

        self._subList.append(subObj)

        return

    def compact(self):
        """
        _compact_

        Compact the subscription list by aggregating the subscriptions where the nodes
        share a list of dataset paths.
        """
        # Bag the subscriptions, keep indexes of bagged items to
        # avoid modifying the list in place or copying the list
        bags = []
        baggedIndexes = set()
        for i, subscriptionA in enumerate(self._subList):
            if i in baggedIndexes:
                continue
            bags.append([subscriptionA])
            for j, subscriptionB in enumerate(self._subList[i + 1:], i + 1):
                if j in baggedIndexes:
                    continue
                if subscriptionA.isEqualOptions(subscriptionB) and \
                    subscriptionA.isEqualDatasetPaths(subscriptionB):
                    bags[-1].append(subscriptionB)
                    baggedIndexes.add(j)

        # Aggregate the subscriptions in the bags
        newSubList = []
        for bag in bags:
            anchorSubscription = bag[0]
            for subscription in bag[1:]:
                anchorSubscription.addNodes(subscription)
            newSubList.append(anchorSubscription)
        self._subList = newSubList

    def getSubscriptionList(self):
        return self._subList



"""
Unit tests for Unified/RequestInfo.py module

Author: Valentin Kuznetsov <vkuznet [AT] gmail [DOT] com>
"""
from __future__ import division, print_function

import unittest

from WMCore.MicroService.MSTransferor.RequestInfo import RequestInfo


class RequestInfoTest(unittest.TestCase):
    "Unit test for RequestInfo module"

    def setUp(self):
        self.mode = 'test'
        self.reqInfo = RequestInfo(microConfig={}, uniConfig={})


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function, division

ALL_ROLES = ['developer', 'admin', 'data-manager', 'production-operator', 'web-service']
ALL_GROUP = ['reqmgr', 'facops', 'dataops']
ADMIN_ROLES = ['admin', 'web-service']
ADMIN_GROUP = ['reqmgr', 'facops']

ADMIN_PERMISSION = {'role': ADMIN_ROLES, 'group': ADMIN_GROUP}
DEFAULT_STATUS_PERMISSION = {'role': ALL_ROLES, 'group': ALL_GROUP}


PERMISSION_BY_REQUEST_TYPE = {
    "ADMIN" : ADMIN_PERMISSION,
    "NOSTATUS":  DEFAULT_STATUS_PERMISSION, # permmission for updating the values without updating status
    "DEFAULT_STATUS": DEFAULT_STATUS_PERMISSION, # permmission for updating the status change
    "DEFAULT_TYPE": {"new": DEFAULT_STATUS_PERMISSION,
                     "assignment-approved": DEFAULT_STATUS_PERMISSION,
                     "assigned": DEFAULT_STATUS_PERMISSION},
    "StoreResult" : {"new": DEFAULT_STATUS_PERMISSION,
                     "assignment-approved": DEFAULT_STATUS_PERMISSION,
                     "assigned": DEFAULT_STATUS_PERMISSION}
    }



"""
_NewSubscription_

MySQL implementation of DBS3Buffer.NewSubscription

Created on May 2, 2013

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter


class NewSubscription(DBFormatter):
    """
    _NewSubscription_

    Create a new subscription in the database
    """

    sql = """INSERT IGNORE INTO dbsbuffer_dataset_subscription
            (dataset_id, site, custodial, auto_approve, move, priority, subscribed, phedex_group, delete_blocks, dataset_lifetime)
            VALUES (:id, :site, :custodial, :auto_approve, :move, :priority, 0, :phedex_group, :delete_blocks, :dataset_lifetime)
          """

    def _createPhEDExSubBinds(self, datasetID, subscriptionInfo, custodialFlag):
        """
        Creates the database binds for both custodial and non custodial data
        placements.

        :param datasetID: integer with the dataset id
        :param subscriptionInfo: dictionary object from the request spec
        :param custodialFlag: boolean flag defining whether it's custodial or not
        :return: a list of dictionary binds
        """
        # NOTE: the subscription information is already validated upstream, at
        # the request factory. Thus, there is no need to cast/validate anything
        # at this level.

        # DeleteFromSource is not supported for move subscriptions
        delete_blocks = None
        if custodialFlag:
            sites = subscriptionInfo['CustodialSites']
            phedex_group = subscriptionInfo['CustodialGroup']
            isMove = 1 if subscriptionInfo['CustodialSubType'] == 'Move' else 0
        else:
            sites = subscriptionInfo['NonCustodialSites']
            phedex_group = subscriptionInfo['NonCustodialGroup']
            isMove = 1 if subscriptionInfo['NonCustodialSubType'] == 'Move' else 0
        if not isMove:
            delete_blocks = 1 if subscriptionInfo.get('DeleteFromSource', False) else None

        binds = []
        for site in sites:
            bind = {'id': datasetID,
                    'site': site,
                    'custodial': 1 if custodialFlag else 0,
                    'auto_approve': 1 if site in subscriptionInfo['AutoApproveSites'] else 0,
                    'move': isMove,
                    'priority': subscriptionInfo['Priority'],
                    'phedex_group': phedex_group,
                    'delete_blocks': delete_blocks}
            if subscriptionInfo['DatasetLifetime'] is not None:
                bind.update(dict(dataset_lifetime=subscriptionInfo['DatasetLifetime']))
            binds.append(bind)
        return binds

    def execute(self, datasetID, subscriptionInfo, conn=None, transaction=False):
        """
        _execute_

        Execute the query and retrieve the subscriptions
        """
        binds = self._createPhEDExSubBinds(datasetID, subscriptionInfo, True)
        binds.extend(self._createPhEDExSubBinds(datasetID, subscriptionInfo, False))

        if not binds:
            return

        self.dbi.processData(self.sql, binds=binds, conn=conn, transaction=transaction)



"""
Define Authentiction roles and groups for the access permission for writing the the database
WMCore.REST.Auth authz_match
"""
from __future__ import print_function, division

from WMCore.ReqMgr.DataStructs.ReqMgrConfigDataCache import ReqMgrConfigDataCache

def getWritePermission(request_args):

    requestType = request_args["RequestType"]
    requestStatus = request_args.get("RequestStatus", None)

    permission_config = ReqMgrConfigDataCache.getConfig("PERMISSION_BY_REQUEST_TYPE")
    default_permission = permission_config["DEFAULT_STATUS"]
    if requestStatus is None:
        return permission_config["NOSTATUS"]
    elif requestType not in permission_config:
        return permission_config["DEFAULT_TYPE"].get(requestStatus, default_permission)
    else:
        return permission_config[requestType].get(requestStatus, default_permission)




#!/usr/bin/env python
"""
_Couchapp_


This is a general unittest for me should I need to create and
test couchapps.  Its utility is probably limited for people who
know what they're doing, and for couchapps that will never need
to be altered
"""
import os
import unittest
import threading

import WMCore.WMBase

from WMCore.WMBS.File                   import File
from WMCore.WMBS.Fileset                import Fileset
from WMCore.WMBS.Workflow               import Workflow
from WMCore.WMBS.Subscription           import Subscription
from WMCore.WMBS.JobGroup               import JobGroup
from WMCore.WMBS.Job                    import Job
from WMCore.DataStructs.Run             import Run
from WMCore.Database.CMSCouch           import CouchServer
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.WMSpec.Makers.TaskMaker     import TaskMaker
from WMCore.Services.UUIDLib               import makeUUID
from WMCore.FwkJobReport.Report         import Report
from WMQuality.TestInitCouchApp         import TestInitCouchApp as TestInit

from WMCore_t.WMSpec_t.TestSpec     import testWorkload

class CouchappTest(unittest.TestCase):

    def setUp(self):
        myThread = threading.currentThread()

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules = ["WMCore.WMBS"],
                                useDefault = False)
        self.databaseName = "couchapp_t_0"


        # Setup config for couch connections
        config = self.testInit.getConfiguration()

        self.testInit.setupCouch(self.databaseName, "WorkloadSummary")
        self.testInit.setupCouch("%s/jobs" % config.JobStateMachine.couchDBName, "JobDump")
        self.testInit.setupCouch("%s/fwjrs" % config.JobStateMachine.couchDBName, "FWJRDump")
        self.testInit.setupCouch(config.JobStateMachine.summaryStatsDBName, "SummaryStats")

        # Create couch server and connect to databases
        self.couchdb      = CouchServer(config.JobStateMachine.couchurl)
        self.jobsdatabase = self.couchdb.connectDatabase("%s/jobs" % config.JobStateMachine.couchDBName)
        self.fwjrdatabase = self.couchdb.connectDatabase("%s/fwjrs" % config.JobStateMachine.couchDBName)
        self.statsumdatabase = self.couchdb.connectDatabase(config.JobStateMachine.summaryStatsDBName)

        # Create changeState
        self.changeState = ChangeState(config)
        self.config      = config

        # Create testDir
        self.testDir = self.testInit.generateWorkDir()

        return

    def tearDown(self):

        self.testInit.clearDatabase(modules = ["WMCore.WMBS"])
        self.testInit.tearDownCouch()
        self.testInit.delWorkDir()
        #self.testInit.tearDownCouch()
        return


    def createWorkload(self, workloadName = 'Test', emulator = True):
        """
        _createTestWorkload_

        Creates a test workload for us to run on, hold the basic necessities.
        """

        workload = testWorkload("Tier1ReReco")

        taskMaker = TaskMaker(workload, os.path.join(self.testDir, 'workloadTest'))
        taskMaker.skipSubscription = True
        taskMaker.processWorkload()

        workload.save(workloadName)

        return workload

    def createTestJobGroup(self, name = "TestWorkthrough",
                           specLocation = "spec.xml", error = False,
                           task = "/TestWorkload/ReReco", nJobs = 10):
        """
        _createTestJobGroup_

        Generate a test WMBS JobGroup with real FWJRs
        """

        myThread = threading.currentThread()

        testWorkflow = Workflow(spec = specLocation, owner = "Simon",
                                name = name, task = task)
        testWorkflow.create()

        testWMBSFileset = Fileset(name = name)
        testWMBSFileset.create()

        testFileA = File(lfn = makeUUID(), size = 1024, events = 10)
        testFileA.addRun(Run(10, *[12312]))
        testFileA.setLocation('malpaquet')

        testFileB = File(lfn = makeUUID(), size = 1024, events = 10)
        testFileB.addRun(Run(10, *[12312]))
        testFileB.setLocation('malpaquet')

        testFileA.create()
        testFileB.create()

        testWMBSFileset.addFile(testFileA)
        testWMBSFileset.addFile(testFileB)
        testWMBSFileset.commit()
        testWMBSFileset.markOpen(0)

        testSubscription = Subscription(fileset = testWMBSFileset,
                                        workflow = testWorkflow)
        testSubscription.create()

        testJobGroup = JobGroup(subscription = testSubscription)
        testJobGroup.create()

        for i in range(0, nJobs):
            testJob = Job(name = makeUUID())
            testJob.addFile(testFileA)
            testJob.addFile(testFileB)
            testJob['retry_count'] = 1
            testJob['retry_max'] = 10
            testJob['mask'].addRunAndLumis(run = 10, lumis = [12312, 12313])
            testJobGroup.add(testJob)

        testJobGroup.commit()

        report = Report()
        if error:
            path   = os.path.join(WMCore.WMBase.getTestBase(),
                                  "WMComponent_t/JobAccountant_t/fwjrs", "badBackfillJobReport.pkl")
        else:
            path = os.path.join(WMCore.WMBase.getTestBase(),
                                "WMComponent_t/JobAccountant_t/fwjrs", "PerformanceReport2.pkl")
        report.load(filename = path)

        self.changeState.propagate(testJobGroup.jobs, 'created', 'new')
        self.changeState.propagate(testJobGroup.jobs, 'executing', 'created')
        self.changeState.propagate(testJobGroup.jobs, 'complete', 'executing')
        for job in testJobGroup.jobs:
            job['fwjr'] = report
        self.changeState.propagate(testJobGroup.jobs, 'jobfailed', 'complete')
        self.changeState.propagate(testJobGroup.jobs, 'retrydone', 'jobfailed')
        self.changeState.propagate(testJobGroup.jobs, 'exhausted', 'retrydone')
        self.changeState.propagate(testJobGroup.jobs, 'cleanout', 'exhausted')

        testSubscription.completeFiles([testFileA, testFileB])

        return testJobGroup


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import unittest

from WMCore.Services.PhEDEx.DataStructs.SubscriptionList import SubscriptionList, PhEDExSubscription

class SubscriptionListTest(unittest.TestCase):
    """
    _SubscriptionListTest_

    Test class for the subscription list data structure
    """

    def setUp(self):
        """
        _setUp_

        Nothing to setup
        """
        pass

    def tearDown(self):
        """
        _tearDown_

        Nothing to tear down
        """
        pass

    def testSubscriptionList(self):
        """
        _SubscriptionList_

        Check that we can organize and agreggate correctly a bunch of different subscriptions
        in standard scenarios
        """
        subList = SubscriptionList()
        # Two GEN datasets subscribed to many sites non-custodially and custodially to one
        subs = []
        genDatasetA = "/DeadlyNeurotoxinOnTestSubjectSim/Run1970-Test-v2/GEN"
        subs.append({"datasetPathList" : genDatasetA, "nodeList" : "T1_US_FNAL",
                     "group" : "dataops", "custodial" : "y", "move" : "y"})
        subs.append({"datasetPathList" : genDatasetA, "nodeList" : "T2_IT_Bari",
                     "group" : "dataops", "request_only" : "n"})
        subs.append({"datasetPathList" : genDatasetA, "nodeList" : "T2_CH_CERN",
                     "group" : "dataops", "request_only" : "n"})
        subs.append({"datasetPathList" : genDatasetA, "nodeList" : "T2_US_Wisconsin",
                     "group" : "dataops"})
        genDatasetB = "/NotEnoughEnergyToLieIn1.1V/Run1970-Potato-v2/GEN"
        subs.append({"datasetPathList" : genDatasetB, "nodeList" : "T1_IT_CNAF",
                     "group" : "dataops", "custodial" : "y", "move" : "y"})
        subs.append({"datasetPathList" : genDatasetB, "nodeList" : "T2_IT_Bari",
                     "group" : "dataops", "request_only" : "n"})
        subs.append({"datasetPathList" : genDatasetB, "nodeList" : "T2_CH_CERN",
                     "group" : "dataops", "request_only" : "n"})
        subs.append({"datasetPathList" : genDatasetB, "nodeList" : "T2_US_Wisconsin",
                     "group" : "dataops"})
        # RECO,DQM,AOD datasets subscribed custodially to 2 sites
        recoDatasetA = '/TestWeightedCubes/Run1970-Test-v2/%s'
        recoDatasetB = '/RepulsiveGel/Run1970-Test-v2/%s'
        subs.append({"datasetPathList" : recoDatasetA % 'AOD' , "nodeList" : "T1_US_FNAL",
                     "group" : "dataops", "custodial" : "y", "move" : "y"})
        subs.append({"datasetPathList" : recoDatasetA % 'DQM' , "nodeList" : "T1_US_FNAL",
                     "group" : "dataops", "custodial" : "y", "move" : "y"})
        subs.append({"datasetPathList" : recoDatasetB % 'AOD' , "nodeList" : "T1_DE_KIT",
                     "group" : "dataops", "custodial" : "y", "move" : "y"})
        subs.append({"datasetPathList" : recoDatasetB % 'DQM' , "nodeList" : "T1_DE_KIT",
                     "group" : "dataops", "custodial" : "y", "move" : "y"})

        for sub in subs:
            phedexSub = PhEDExSubscription(**sub)
            subList.addSubscription(phedexSub)

        # One subscription per node
        self.assertEqual(len(subList.getSubscriptionList()), 6)
        goldenDatasetLists = [set([genDatasetA, genDatasetB]),
                              set([genDatasetA, recoDatasetA % 'AOD',
                                   recoDatasetA % 'DQM']), set([genDatasetB]),
                              set([recoDatasetB % 'AOD', recoDatasetB % 'DQM'])]
        for sub in subList.getSubscriptionList():
            self.assertTrue(set(sub.getDatasetPaths()) in goldenDatasetLists)

        subList.compact()

        # Compact should have reduced 2 of them to 1
        goldenNodeLists = [set(["T1_US_FNAL"]), set(["T2_IT_Bari", "T2_CH_CERN"]),
                           set(["T1_IT_CNAF"]), set(["T1_DE_KIT"]), set(["T2_US_Wisconsin"])]
        self.assertEqual(len(subList.getSubscriptionList()), 5)
        for sub in subList.getSubscriptionList():
            self.assertTrue(set(sub.getNodes()) in goldenNodeLists)
        return

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

from __future__ import print_function

import json
import os
import unittest

from WMCore_t.ReqMgr_t.TestConfig import config

from WMCore.REST.Test import fake_authz_headers
from WMCore.ReqMgr.Auth import getWritePermission
from WMQuality.REST.RESTBaseUnitTestWithDBBackend import RESTBaseUnitTestWithDBBackend

req_args = {"RequestType": "ReReco", "RequestStatus": None}
ADMIN_PERMISSION = getWritePermission(req_args)

req_args = {"RequestType": "ReReco", "RequestStatus": "completed"}
DEFAULT_STATUS_PERMISSION = getWritePermission(req_args)

req_args = {"RequestType": "ReReco", "RequestStatus": "new"}
CREATE_PERMISSION = getWritePermission(req_args)

DEFAULT_PERMISSION = DEFAULT_STATUS_PERMISSION

req_args = {"RequestType": "ReReco", "RequestStatus": "assinged"}
ASSIGN_PERMISSION = getWritePermission(req_args)

# this needs to move in better location
def insertDataToCouch(couchUrl, couchDBName, data):
    import WMCore.Database.CMSCouch
    server = WMCore.Database.CMSCouch.CouchServer(couchUrl)
    database = server.connectDatabase(couchDBName)

    doc = database.commit(data)
    return doc

def getAuthHeader(hmacData, reqAuth):
    roles = {}
    for role in reqAuth['role']:
        roles[role] = {'group': reqAuth['group']}

    return fake_authz_headers(hmacData, dn = "/TEST/DN/CN", roles = roles, format = "dict")


class ReqMgrTest(RESTBaseUnitTestWithDBBackend):
    """
    Test WorkQueue Service client
    It will start WorkQueue RESTService
    Server DB sets from environment variable.
    Client DB sets from environment variable.

    This checks whether DS call makes without error and return the results.
    Not the correctness of functions. That will be tested in different module.

    """
    def setFakeDN(self):
        # put into ReqMgr auxiliary database under "software" document scram/cmsms
        # which we'll need a little for request injection
        #Warning: this assumes the same structure in jenkins wmcore_root/test
        self.admin_header = getAuthHeader(self.test_authz_key.data, ADMIN_PERMISSION)
        self.create_header = getAuthHeader(self.test_authz_key.data, CREATE_PERMISSION)
        self.default_header = getAuthHeader(self.test_authz_key.data, DEFAULT_PERMISSION)
        self.assign_header = getAuthHeader(self.test_authz_key.data, ASSIGN_PERMISSION)
        self.default_status_header = getAuthHeader(self.test_authz_key.data, DEFAULT_STATUS_PERMISSION)

    def setUp(self):
        self.setConfig(config)
        self.setCouchDBs([(config.views.data.couch_reqmgr_db, "ReqMgr"),
                          (config.views.data.couch_reqmgr_aux_db, None)])
        self.setSchemaModules([])

        RESTBaseUnitTestWithDBBackend.setUp(self)

        self.setFakeDN()
        #print "%s" % self.test_authz_key.data
        self.default_status_header = getAuthHeader(self.test_authz_key.data, DEFAULT_STATUS_PERMISSION)

        normPath = os.path.normpath(os.path.join(os.path.dirname(__file__), '..', '..', '..', '..'))
        rerecoPath = os.path.join(normPath, 'data/ReqMgr/requests/DMWM/ReReco_RunBlockWhite.json')
        with open(rerecoPath) as jObj:
            rerecoArgs = json.load(jObj)
        self.rerecoCreateArgs = rerecoArgs["createRequest"]
        self.rerecoCreateArgs["PrepID"] = "test_prepid"
        self.rerecoAssignArgs = rerecoArgs["assignRequest"]
        # overwrite rereco args
        self.rerecoAssignArgs["AcquisitionEra"] = "test_aqc"

        taskChainPath = os.path.join(normPath, 'data/ReqMgr/requests/DMWM/TaskChain_InclParents.json')
        with open(taskChainPath) as jObj:
            lheArgs = json.load(jObj)
        self.lheStep0CreateArgs = lheArgs["createRequest"]
        self.lheStep0AssignArgs = lheArgs["assignRequest"]
        self.lheStep0AssignArgs["AcquisitionEra"] = "test_aqc"

        cmsswDoc = {"_id": "software"}
        cmsswDoc[self.rerecoCreateArgs["ScramArch"]] =  []
        cmsswDoc[self.rerecoCreateArgs["ScramArch"]].append(self.rerecoCreateArgs["CMSSWVersion"])
        insertDataToCouch(os.getenv("COUCHURL"), config.views.data.couch_reqmgr_aux_db, cmsswDoc)


    def tearDown(self):
        RESTBaseUnitTestWithDBBackend.tearDown(self)


    def getRequestWithNoStale(self, query):
        prefixWithNoStale = "data/request?_nostale=true&"
        return self.jsonSender.get(prefixWithNoStale + query,
                                   incoming_headers=self.default_header)

    def postRequestWithAuth(self, data):
        return self.jsonSender.post('data/request', data, incoming_headers=self.create_header)

    def putRequestWithAuth(self, requestName, data):
        """
        WMCore.REST doesn take query for the put request.
        data need to send on the body
        """
        return self.jsonSender.put('data/request/%s' % requestName, data,
                                     incoming_headers=self.assign_header)

    def getMultiRequestsWithAuth(self, data):
        return self.jsonSender.post('data/request/bynames', data, incoming_headers=self.create_header)

    def cloneRequestWithAuth(self, requestName, params = {}):
        """
        WMCore.REST doesn take query for the put request.
        data need to send on the body
        """
        params["OriginalRequestName"] = requestName
        return self.jsonSender.put('data/request/clone', params,
                                     incoming_headers=self.assign_header)

    def resultLength(self, response, format="dict"):
        # result is dict format
        if format == "dict":
            return len(response[0]['result'][0])
        elif format == "list":
            return  len(response[0]['result'])

    def insertRequest(self, args):
        # test post method
        response = self.postRequestWithAuth(self.rerecoCreateArgs)
        self.assertEqual(response[1], 200)
        requestName = response[0]['result'][0]['request']
        return requestName

    def testRequestSimpleCycle(self):
        """
        test request cycle with one request without composite get condition.
        post, get, put
        """

        # test post method
        requestName = self.insertRequest(self.rerecoCreateArgs)

        ## test get method
        # get by name
        response = self.getRequestWithNoStale('name=%s' % requestName)
        self.assertEqual(response[1], 200, "get by name")
        self.assertEqual(self.resultLength(response), 1)

        # get by status
        response = self.getRequestWithNoStale('status=new')
        self.assertEqual(response[1], 200, "get by status")
        self.assertEqual(self.resultLength(response), 1)

        #this create cache
        # need to find the way to reste Etag or not getting from the cache
#         response = self.getRequestWithNoStale('status=assigned')
#         self.assertEqual(response[1], 200, "get by status")
#         self.assertEqual(self.resultLength(response), 0)

        # get by prepID
        response = self.getRequestWithNoStale('prep_id=%s' % self.rerecoCreateArgs["PrepID"])
        self.assertEqual(response[1], 200)
        self.assertEqual(self.resultLength(response), 1)
        #import pdb
        #pdb.set_trace()
        response = self.getRequestWithNoStale('campaign=%s' % self.rerecoCreateArgs["Campaign"])
        self.assertEqual(response[1], 200)
        self.assertEqual(self.resultLength(response), 1)

        response = self.getRequestWithNoStale('inputdataset=%s' % self.rerecoCreateArgs["InputDataset"])
        self.assertEqual(response[1], 200)
        self.assertEqual(self.resultLength(response), 1)

        response = self.getRequestWithNoStale('mc_pileup=%s' % self.rerecoCreateArgs["MCPileup"])
        self.assertEqual(response[1], 200)
        self.assertEqual(self.resultLength(response), 1)

        response = self.getRequestWithNoStale('data_pileup=%s' % self.rerecoCreateArgs["DataPileup"])
        self.assertEqual(response[1], 200)
        self.assertEqual(self.resultLength(response), 1)


        # test put request with just status change
        data = {'RequestStatus': 'assignment-approved'}
        self.putRequestWithAuth(requestName, data)
        response = self.getRequestWithNoStale('status=assignment-approved')
        self.assertEqual(response[1], 200, "put request status change")
        self.assertEqual(self.resultLength(response), 1)

        # assign with team
        # test put request with just status change
        data = {'RequestStatus': 'assigned'}
        data.update(self.rerecoAssignArgs)
        self.putRequestWithAuth(requestName, data)
        response = self.getRequestWithNoStale('status=assigned')
        self.assertEqual(response[1], 200, "put request status change")
        self.assertEqual(self.resultLength(response), 1)

        response = self.getRequestWithNoStale('status=assigned&team=%s' %
                                             self.rerecoAssignArgs['Team'])
        self.assertEqual(response[1], 200, "put request status change")
        self.assertEqual(self.resultLength(response), 1)

        response = self.getMultiRequestsWithAuth([requestName])
        self.assertEqual(self.resultLength(response), 1)
        self.assertEqual(list(response[0]['result'][0])[0], requestName)

        #response = self.cloneRequestWithAuth(requestName)
        #self.assertEqual(response[1], 200, "put request clone")
        #response = self.getRequestWithNoStale('status=new')
        #self.assertEqual(self.resultLength(response), 1)

    def atestRequestCombinedGetCall(self):
        """
        test request composite get call
        """

        # test post method
        from pprint import pprint
        del self.rerecoCreateArgs["CMSSWVersion"]
        pprint(self.rerecoCreateArgs)
        rerecoReqName = self.insertRequest(self.rerecoCreateArgs)
        lheReqName = self.insertRequest(self.lheStep0CreateArgs)
        ## test get method
        # get by name
        response = self.getRequestWithNoStale('name=%s&name=%s' % (rerecoReqName, lheReqName))
        self.assertEqual(response[1], 200, "get by name")
        self.assertEqual(self.resultLength(response), 2)

        # get by status
        response = self.getRequestWithNoStale('status=new')
        self.assertEqual(response[1], 200, "get by status")
        self.assertEqual(self.resultLength(response), 2)

        # get by prepID
        response = self.getRequestWithNoStale('prep_id=%s' % self.rerecoCreateArgs["PrepID"])
        self.assertEqual(response[1], 200)
        self.assertEqual(self.resultLength(response), 2)
        #import pdb
        #pdb.set_trace()
        response = self.getRequestWithNoStale('campaign=%s' % self.rerecoCreateArgs["Campaign"])
        self.assertEqual(response[1], 200)
        self.assertEqual(self.resultLength(response), 2)


    def atestRequestClone(self):
        requestName = self.insertRequest(self.rerecoCreateArgs)
        response = self.cloneRequestWithAuth(requestName)
        print(response)
        self.assertEqual(response[1], 200, "put request clone")
        response = self.getRequestWithNoStale('status=new')
        self.assertEqual(self.resultLength(response), 2)

if __name__ == '__main__':
    unittest.main()
